/*
 * Copyright (c) 2010 Adrian Chadd, Xenion Pty Ltd
 * Copyright (c) 2008 Sam Leffler, Errno Consulting
 * Copyright (c) 2008 Atheros Communications, Inc.
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * $FreeBSD$
 */
#include "opt_ah.h"

#include "ah.h"
#include "ah_internal.h"

#include "ar5416/ar5416.h"
#include "ar5416/ar5416reg.h"
#include "ar5416/ar5416phy.h"

#include "ar5416/ar5416_hwops.h"

#define	AH_WAIT_TIMEOUT		100000	/* (us) */
#define	BASE_ACTIVATE_DELAY	100	/* 100 usec */

/* Defined for now in ar5416_reset.c */
extern void ar5416InitBB(struct ath_hal *ah, const struct ieee80211_channel *);
extern void ar5416InitPLL(struct ath_hal *ah, const struct ieee80211_channel *);
extern void ar5416SetDeltaSlope(struct ath_hal *, const struct ieee80211_channel *);

static void
ar5008_hw_set_nf_limits(struct ath_hal *ah)
{
	AH5416(ah)->nf_2g.max = AR_PHY_CCA_MAX_GOOD_VAL_5416_2GHZ;
	AH5416(ah)->nf_2g.min = AR_PHY_CCA_MIN_GOOD_VAL_5416_2GHZ;
	AH5416(ah)->nf_2g.nominal = AR_PHY_CCA_NOM_VAL_5416_2GHZ;
	AH5416(ah)->nf_5g.max = AR_PHY_CCA_MAX_GOOD_VAL_5416_5GHZ;
	AH5416(ah)->nf_5g.min = AR_PHY_CCA_MIN_GOOD_VAL_5416_5GHZ;
	AH5416(ah)->nf_5g.nominal = AR_PHY_CCA_NOM_VAL_5416_5GHZ;
}

static uint32_t
ar5416_hw_compute_pll_control(struct ath_hal *ah, const struct ieee80211_channel *chan)
{
	uint32_t pll;

	pll = AR_RTC_PLL_REFDIV_5 | AR_RTC_PLL_DIV2;
	if (chan != AH_NULL) {
		if (IEEE80211_IS_CHAN_HALF(chan))
			pll |= SM(0x1, AR_RTC_PLL_CLKSEL);
		else if (IEEE80211_IS_CHAN_QUARTER(chan))
			pll |= SM(0x2, AR_RTC_PLL_CLKSEL);
		else if (IEEE80211_IS_CHAN_5GHZ(chan))
			pll |= SM(0xa, AR_RTC_PLL_DIV);
		else
			pll |= SM(0xb, AR_RTC_PLL_DIV);
	} else
		pll |= SM(0xb, AR_RTC_PLL_DIV);  

	return pll;
}

#ifdef ADRIAN_NOTYET
/*
 * This isn't used yet - it's just a placeholder for when/if I migrate to using it.
 * The current NF read code is already indirected through the HAL.
 */
static void
ar5008_hw_do_getnf(struct ath_hal *ah, int16_t nfarray[NUM_NOISEFLOOR_READINGS])
{
	int16_t nf;

	nf = MS(OS_REG_READ(ah, AR_PHY_CCA), AR_PHY_MINCCA_PWR);
	nfarray[0] = sign_extend(nf, 9);

	nf = MS(OS_REG_READ(ah, AR_PHY_CH1_CCA), AR_PHY_CH1_MINCCA_PWR);
	nfarray[1] = sign_extend(nf, 9);

	nf = MS(OS_REG_READ(ah, AR_PHY_CH2_CCA), AR_PHY_CH2_MINCCA_PWR);
	nfarray[2] = sign_extend(nf, 9);

	if (!IEEE80211_IS_CHAN_HT40(AH_PRIVATE(ah)->ah_curchan))
		return;

	nf = MS(OS_REG_READ(ah, AR_PHY_EXT_CCA), AR_PHY_EXT_MINCCA_PWR);
	nfarray[3] = sign_extend(nf, 9);

	nf = MS(OS_REG_READ(ah, AR_PHY_CH1_EXT_CCA), AR_PHY_CH1_EXT_MINCCA_PWR);
	nfarray[4] = sign_extend(nf, 9);

	nf = MS(OS_REG_READ(ah, AR_PHY_CH2_EXT_CCA), AR_PHY_CH2_EXT_MINCCA_PWR);
	nfarray[5] = sign_extend(nf, 9);
}
#endif

/*
 * This is the HAL version; not the ath9k version
 */
static void
ar5416_hw_set_rfmode(struct ath_hal *ah, const struct ieee80211_channel *chan)
{
 	uint32_t rfMode;

	if (chan == AH_NULL)
		return;

	/* treat channel B as channel G , no  B mode suport in owl */
	rfMode = IEEE80211_IS_CHAN_CCK(chan) ? AR_PHY_MODE_DYNAMIC : AR_PHY_MODE_OFDM;
	if (AR_SREV_MERLIN_20(ah) && IS_5GHZ_FAST_CLOCK_EN(ah, chan)) {
		/* phy mode bits for 5GHz channels require Fast Clock */
		rfMode |= AR_PHY_MODE_DYNAMIC | AR_PHY_MODE_DYN_CCK_DISABLE;
	} else if (!AR_SREV_MERLIN_10_OR_LATER(ah)) {
		rfMode |= IEEE80211_IS_CHAN_5GHZ(chan) ? AR_PHY_MODE_RF5GHZ : AR_PHY_MODE_RF2GHZ;
	}
	OS_REG_WRITE(ah, AR_PHY_MODE, rfMode);
}

static HAL_BOOL
ar5008_hw_rfbus_req(struct ath_hal *ah)
{
	OS_REG_WRITE(ah, AR_PHY_RFBUS_REQ, AR_PHY_RFBUS_REQ_EN);
        return ath_hal_waitfor(ah, AR_PHY_RFBUS_GRANT, AR_PHY_RFBUS_GRANT_EN,
	    AR_PHY_RFBUS_GRANT_EN, AH_WAIT_TIMEOUT);
}

static void
ar5008_hw_rfbus_done(struct ath_hal *ah)
{
	uint32_t synthDelay = OS_REG_READ(ah, AR_PHY_RX_DELAY) & AR_PHY_RX_DELAY_DELAY;

	if (IEEE80211_IS_CHAN_B(AH_PRIVATE(ah)->ah_curchan))
		synthDelay = (4 * synthDelay) / 22;
	else
		synthDelay /= 10;
	OS_DELAY(synthDelay + BASE_ACTIVATE_DELAY);
	OS_REG_WRITE(ah, AR_PHY_RFBUS_REQ, 0);
}

static void
ar5008_hw_mark_phy_inactive(struct ath_hal *ah)
{
	OS_REG_WRITE(ah, AR_PHY_ACTIVE, AR_PHY_ACTIVE_DIS);
}

static void
ar5008_hw_olc_init(struct ath_hal *ah)
{
}

static void
ar5008_hw_restore_chainmask(struct ath_hal *ah)
{
	int rx_chainmask = AH5416(ah)->ah_rx_chainmask;

	if ((rx_chainmask == 0x5) || (rx_chainmask == 0x3)) {
		OS_REG_WRITE(ah, AR_PHY_RX_CHAINMASK, rx_chainmask);
		OS_REG_WRITE(ah, AR_PHY_CAL_CHAINMASK, rx_chainmask);
		OS_REG_WRITE(ah, AR_PHY_RX_CHAINMASK, rx_chainmask);
		OS_REG_WRITE(ah, AR_PHY_CAL_CHAINMASK, rx_chainmask);
	}
}

void
ar5416_hwops_setup(struct ath_hal *ah)
{
	const uint32_t ar5416_cca_regs[6] = {
		AR_PHY_CCA,
		AR_PHY_CH1_CCA,
		AR_PHY_CH2_CCA,
		AR_PHY_EXT_CCA,
		AR_PHY_CH1_EXT_CCA,
		AR_PHY_CH2_EXT_CCA
	};

	/* the ath9k phyops list */

	/* rf_set_freq - phy op setChannel, via ar5212SetChannel() */

	/* spur_mitigate_freq - HAL spurMitigate */

	/* rf_alloc_ext_banks */
	/* rf_free_ext_banks */
	/* set_rf_regs - phy op setRfRegs */

	/* set_channel_regs */

	/* init_bb */
	AH5416(ah)->ah_init_bb = ar5416InitBB;

	/* process_ini - HAL ah_writeIni */

	/* olc_init */
	AH5416(ah)->ah_olc_init = ar5008_hw_olc_init;

	/* set_rfmode */
	AH5416(ah)->ah_set_rfmode = ar5416_hw_set_rfmode;

	/* mark_phy_inactive */
	AH5416(ah)->ah_mark_phy_inactive = ar5008_hw_mark_phy_inactive;

	/* set_delta_slope */
	AH5416(ah)->ah_set_delta_slope = ar5416SetDeltaSlope;

	/* rfbus_req */
	AH5416(ah)->ah_rfbus_req = ar5008_hw_rfbus_req;

	/* rfbus_done */
	AH5416(ah)->ah_rfbus_done = ar5008_hw_rfbus_done;

	/* enable_rfkill */
	AH5416(ah)->ah_enable_rfkill = ar5212EnableRfKill;

	/* restore_chainmask */
	AH5416(ah)->ah_restore_chainmask = ar5008_hw_restore_chainmask;

	/* set_diversity - via HAL setCapability(HAL_CAP_DIVERSITY) */

	/* ani_control */
	/* ani_cache_ini_regs */

	AH5416(ah)->ah_compute_pll_control = ar5416_hw_compute_pll_control;
	AH5416(ah)->ah_set_nf_limits = ar5008_hw_set_nf_limits;

	/* my local phy ops */
#ifdef ADRIAN_NOTYET
	/*
	 * Disable this for now - or I'll forget to override it when
	 * it comes time to test this versus AR92xx
	 *  - adrian
	 */
	AH5416(ah)->ah_do_getnf = ar5008_hw_do_getnf;
#endif
	AH5416(ah)->ah_init_pll = ar5416InitPLL;

	/* the ath9k hwops list */

	/* config_pci_powersave */
	/* rx_enable */
	/* set_desc_link */
	/* get_desc_link */

	/* calibrate */
	AH5416(ah)->ah_init_calibrate = ar5416InitCal;

	/* get_isr */
	/* fill_txdesc */
	/* proc_txdesc */
	/* set11n_txdesc */
	/* set11n_ratescenario */
	/* set11n_aggr_first */
	/* set11n_aggr_middle */
	/* set11n_aggr_last */
	/* clr11n_aggr */
	/* set11n_burstduration */
	/* set11n_virtualmorefrag */

	/* ani_proc_mib_event - in HAL */
	/* ani_monitor - in HAL */

	ar5008_hw_set_nf_limits(ah);
	memcpy(AH5416(ah)->nf_regs, ar5416_cca_regs, sizeof(AH5416(ah)->nf_regs));
}
