/*
 * Copyright (c) 2010 Adrian Chadd, Xenion Pty Ltd
 * Copyright (c) 2008 Sam Leffler, Errno Consulting
 * Copyright (c) 2008 Atheros Communications, Inc.
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * $FreeBSD$
 */
#include "opt_ah.h"

#include "ah.h"
#include "ah_internal.h"
#include "ah_devid.h"

#include "ar5416/ar5416.h"
#include "ar5416/ar5416reg.h"
#include "ar5416/ar5416_hwops.h"

#include "ar9002/ar9280.h"
#include "ar9002/ar9280_phy.h"
#include "ar9002/ar9280_hwops.h"

static void
ar9280_hw_set_nf_limits(struct ath_hal *ah)
{
	AH5416(ah)->nf_2g.max = AR_PHY_CCA_MAX_GOOD_VAL_9280_2GHZ;
	AH5416(ah)->nf_2g.min = AR_PHY_CCA_MIN_GOOD_VAL_9280_2GHZ;
	AH5416(ah)->nf_2g.nominal = AR_PHY_CCA_NOM_VAL_9280_2GHZ;
	AH5416(ah)->nf_5g.max = AR_PHY_CCA_MAX_GOOD_VAL_9280_5GHZ;
	AH5416(ah)->nf_5g.min = AR_PHY_CCA_MIN_GOOD_VAL_9280_5GHZ;
	AH5416(ah)->nf_5g.nominal = AR_PHY_CCA_NOM_VAL_9280_5GHZ;
}

uint32_t
ar9280_hw_compute_pll_control(struct ath_hal *ah, const struct ieee80211_channel *chan)
{
	uint32_t pll;

	if (AR_SREV_MERLIN_20(ah) && chan != AH_NULL && IEEE80211_IS_CHAN_5GHZ(chan)) {
		/*
		 * PLL WAR for Merlin 2.0/2.1
		 * When doing fast clock, set PLL to 0x142c
		 * Else, set PLL to 0x2850 to prevent reset-to-reset variation
		 */
		pll = IS_5GHZ_FAST_CLOCK_EN(ah, chan) ? 0x142c : 0x2850;
	} else {
		pll = SM(0x5, AR_RTC_SOWL_PLL_REFDIV);
		if (chan != AH_NULL) {
			if (IEEE80211_IS_CHAN_HALF(chan))
				pll |= SM(0x1, AR_RTC_SOWL_PLL_CLKSEL);
			else if (IEEE80211_IS_CHAN_QUARTER(chan))
				pll |= SM(0x2, AR_RTC_SOWL_PLL_CLKSEL);
			else if (IEEE80211_IS_CHAN_5GHZ(chan))
				pll |= SM(0x28, AR_RTC_SOWL_PLL_DIV);
			else
				pll |= SM(0x2c, AR_RTC_SOWL_PLL_DIV);
		} else
			pll |= SM(0x2c, AR_RTC_SOWL_PLL_DIV);
	}
	return pll;
}

/*
 * Involved in temperature calibration - not yet written/ported!
 *
 * XXX don't forget to go through the EEPROM code in ath9k to
 * XXX make sure the initial power/gain tables are setup.
 *
 * - adrian
 */
static void
ar9280_hw_olc_init(struct ath_hal *ah)
{
	int i;
	struct ath_hal_9280 *ahp = AH9280(ah);

	if (OLC_FOR_AR9280_20_LATER)
		return;

	for (i = 0; i < AR9280_TX_GAIN_TABLE_SIZE; i++)
		ahp->originalGain[i] = MS(OS_REG_READ(ah, AR_PHY_TX_GAIN_TBL1 + i * 4),
		    AR_PHY_TX_GAIN);
		    ahp->PDADCdelta = 0;
}

/*
 * Start with the AR5416 ops; then override ar9280-specific stuff
 */
void
ar9280_hwops_setup(struct ath_hal *ah)
{
	ar5416_hwops_setup(ah);

	/* ar9002-overriden stuff from ath9k */
	/* set_rf_regs */
	/* rf_alloc_ext_banks */
	/* rf_free_ext_banks */
	/* rf_set_freq */
	/* spur_mitigate_freq */
	/* olc_init */
	AH5416(ah)->ah_olc_init = ar9280_hw_olc_init;

	/* compute_pll_control */
	AH5416(ah)->ah_compute_pll_control = ar9280_hw_compute_pll_control;

	/* do_getnf */

	/* My stuff */
	AH5416(ah)->ah_set_nf_limits = ar9280_hw_set_nf_limits;
	ar9280_hw_set_nf_limits(ah);
}
